from __future__ import print_function
from readers import readers
from summary import summary
from stats import stats
from report import report
from plot import plot
from detrend import detrend
from autoreg import ar_func, autoreg
from chron import chron
from chron_stabilized import chron_stabilized
from xdate import xdate
from series_corr import series_corr
from writers import write


__copyright__ = """
   dplPy for tree ring width time series analyses
   Copyright (C) 2024  OpenDendro

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""
__license__ = "GNU GPLv3"

#!/usr/bin/python
# -*- coding: utf-8 -*-

# Date: 5/27/2022
# Author: Tyson Lee Swetnam
# Project: OpenDendro dplPy
# Description: Imports main functionality for package
# example usage from Python Console: 
# >>> import dplpy as dpl
# >>> dpl.readme()
# >>> dpl.help()
# >>> dpl.readers(input="tests/csv/ca533.csv")
# >>> dpl.writers(input="tests/csv/ca533.csv",output="ca533.rwl")

import argparse
import webbrowser

# Help Menu
def help():
    try:
        print("*Welcome to the dplPy Help Menu*")
        print("")
        print("....:....⋮....:....⋮....:....⋮....:....⋮....:....⋮....:....⋮....:....⋮....:....⋮....:....⋮....:....⁞")
        print("")
        print("README \n")
        print("to view the documentation online from terminal, type: \n")
        print("$ python src/dplpy.py readme() \n")
        print("from Python Console:")
        print(">>> import dplpy")
        print(">>> dplpy.readme() \n")
        print("or visit our website click url: https://opendendro.org/python \n")
        print("READERS \n")
        print("Import ring width series in a terminal: \n")
        print("$ python src/dplpy.py reader --input /folder/filename.csv \n")
        print(
            "arguments: \n"
        )
        print(
            " --input, -i : single ring width series formatted file (.CSV, .RWL, .TXT) \n"
            " --help, -h : echo the help menu \n"
            )
        print("from the Python Console: \n")
        print(">>> import dplpy as dpl")
        print(">>> dpl.readers(\"/folder/filename.csv\") \n")
        print("WRITERS \n")
        print("Write or convert outputs to new file from terminal, type: \n")
        print("$ python src/dplpy.py writer --input /folder/in_filename.csv --output /folder/out_filename.rwl \n")
        print("arguments: \n")
        print(
            " --input, -i : single ring width series formatted file (.CSV, .RWL, .TXT) \n"
            " --output, -i : single ring with series formatted file (.CSV, .RWL, .TXT \n"
            " --help, -h : echo the help menu \n"
        )
        print("from the Python Console:")
        print("")
        print(">>> import dplpy as dpl")
        print(">>> dpl.writers(input=\"/folder/filename.csv\",output=\"/outputfolder/outputfile.rwl\") \n")
        print("SUMMARY STATISTICS")
        print("TBD")
        print("")
        print("")
        print("END HELP MANUAL")
    except Exception as e:
        print(e)
# set the definition for the help function
def help_from_parser(args):
    help()

# Open the Website README (Manual documentation) 
def readme():
    try:
        a = webbrowser.open("https://opendendro.org/dplpy-man/", new=2)
        print("Success: Check your web browser for a new tab")
        if a is False:
            print("Your computer does not use a monitor, and cannot display the webpages")
    except Exception as e:
        print(e)
# set the definition for the help function
def readme_from_parser(args):
    readme()

# Get package versioning -- commented out until we add dplPy to pypi.org
#def dplpy_version():
#    url = "https://pypi.org/project/dplpy/"
#    source = requests.get(url)
#    html_content = source.text
#    soup = BeautifulSoup(html_content, "html.parser")
#    company = soup.find("h1")
#    vcheck = ob1.compareVersion(
#        company.string.strip().split(" ")[-1],
#        pkg_resources.get_distribution("dplpy").version,
#    )
#    if vcheck == 1:
#        print(
#            "\n"
#            + "========================================================================="
#        )
#        print(
#            "Current version of dplPy is {} upgrade to lastest version: {}".format(
#                pkg_resources.get_distribution("dplpy").version,
#                company.string.strip().split(" ")[-1],
#            )
#        )
#        print(
#            "========================================================================="
#        )
#    elif vcheck == -1:
#        print(
#            "\n"
#            + "========================================================================="
#        )
#        print(
#            "Possibly running staging code {} compared to pypi release {}".format(
#                pkg_resources.get_distribution("dplpy").version,
#                company.string.strip().split(" ")[-1],
#            )
#        )
#        print(
#            "========================================================================="
#        )
#
#dplpy_version()



# set the definition for the Readers functions (from readers.py)
def readers_from_parser(args):
    readers(input=args.input)

# set the definition for the Writers funcitons (from writers.py)
# def writers_from_parser(args):
#    writers(input=args.input,output=args.output)

# Summary
def summary_from_parser(args):
    summary(input=args.input)

# Generate summary statistics
def stats_from_parser(args):
    stats(input=args.input)

def report_from_parser(args):
    report(input=args.input)

def plot_from_parser(args):
    plot(input=args.input)

def chron_from_parser(args):
    chron(input=args.input)

def detrend_from_parser(args):
    detrend(input=args.input)

def ar_func_from_parser(args):
    ar_func(input=args.input)

def autoreg_from_parser(args):
    autoreg(input=args.input)

def xdate_from_parser(args):
    xdate(input=args.input)

def series_corr_from_parser(args):
    series_corr(input=args.input)

def chron_stabilized_from_parser(args):
    chron_stabilized(input=args.input)

def write_from_parser(args):
    write(input=args.input)

# creates whitespace
print("")

spacing = "                               "

# Import definition functions from other files in the src/ path

def main(args=None):
    parser = argparse.ArgumentParser(description="dplPy v0.1") # update version as we update packages
    subparsers = parser.add_subparsers()

# Help Documentation parser
    parser_help = subparsers.add_parser(
        "help", help="Echo the Help Menu"
    )

    parser_help.set_defaults(func=help_from_parser)

# README Documentation pages parser
    parser_readme = subparsers.add_parser(
        "readme", help="Goes to the website: https://opendendro.org/python/"
    )

    parser_readme.set_defaults(func=readme_from_parser)

# Readers file input parser
    parser_readers = subparsers.add_parser(
        "readers", help="Read input ring width series files (.CSV, .JSON, .RWL, .TXT) into an array"
    )

    parser_readers.add_argument(
        "--input", "-i",
        help="<Required> select a valid input ring width series file (.CSV, .JSON, .RWL, .TXT) ",
        required=True
    )

    parser_readers.set_defaults(func=readers_from_parser)

# Writers file output parser
#    parser_writers = subparsers.add_parser(
#        "writer", help="Write out ring width series from array to file (.CSV, .JSON, .RWL, .TXT)"
#    )
#    parser_writers.add_argument(
#        "--input", "-i",
#        help="<Required> select a valid input ring width series file (.CSV, .JSON, .RWL, .TXT) ",
#        required=True
#    )
#    parser_writers.add_argument(
#        "--output", "-o",
#        help="<Required> select a valid output ring width series file (.CSV, .JSON, .RWL, .TXT) ",
#        required=True
#    )
#    parser_writers.set_defaults(func=writers_from_parser)

# Summary Statistics parser
#    parser_summary = subparsers.add_parser(
#        "summary", help="Prints out the summary statistics for an array or input file"
#    )
#    parser_writers.add_argument(
#        "--input", "-i",
#        help="<Required> select a valid input ring width series file (.CSV, .JSON, .RWL, .TXT) ",
#        required=True
#    )
#    parser_summary.add_argument(
#       "--stats", "-s",
#       help="<Default> calculates a specific set of summary statistics from the input file series",
#       required=False
#    )
#
#    parser_summary.set_defaults(func=read_from_parser)

# TBD

# finish argument parsing
    args = parser.parse_args()

    try:
        func = args.func
    except AttributeError:
        parser.error("too few arguments")
    func(args)

if __name__ == "__main__":
    main()


__copyright__ = """
   dplPy for tree ring width time series analyses
   Copyright (C) 2024  OpenDendro

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

__license__ = "GNU GPLv3"

#!/usr/bin/python
# -*- coding: utf-8 -*-

# Date: 11/1/2022
# Author: Ifeoluwa Ale
# Title: autoreg.py
# Description: Contains methods that fit series to autoregressive models and perform functions 
#              related to AR modeling.
#              NOTE: This function only accepts pandas series and dataframes as parameters.

from statsmodels.tsa.ar_model import ar_select_order
import pandas as pd
import numpy as np
import warnings

def ar_func(data, max_lag=5):
    """Auto Regressive (AR) functions 
      
    Extended Summary
    ---------------
    Contains methods that fit series to autoregressive models and perform functions 
    related to AR modeling.
  
    NOTE: This function only accepts pandas series and dataframes as parameters. 
   
    Parameters
    ----------
    data :  str
            a data file (.CSV or .RWL) or a pandas dataframe imported from dpl.readers().
    series: str
            an individual series within a chronology `data` file.
    lag:    int, default 5
            nuber of years.
   
    Returns
    -------
    data :  pandas dataframe
    
    Examples
    --------
    >>> import dplpy as dpl 
    >>> data = dpl.readers("../tests/data/csv/file.csv")
    >>> dpl.ar_func(data['series name']) -> returns residuals plus mean of best fit 
                                            from AR models with max lag of either 5 
                                            (default) or specified number
    
    References
    ----------
    .. [1] https:/opendendro.org/dplpy-man/#ar_func 
    
    """
    if isinstance(data, pd.DataFrame):
        start_df = pd.DataFrame(index=pd.Index(data.index))
        to_concat = [start_df]
        for column in data.columns:
            to_concat.append(ar_func_series(data[column], max_lag))
        res = pd.concat(to_concat, axis=1)
        return res
    elif isinstance(data, pd.Series):
        res = ar_func_series(data, max_lag)
        return res
    else:
        raise TypeError("Data argument should be either pandas dataframe or pandas series.")

# This function returns residuals plus mean of the best fit AR
# model of the data
def ar_func_series(data, max_lag):
    nullremoved_data = data.dropna()
    pars = autoreg(nullremoved_data, max_lag)
    
    y = nullremoved_data
    
    yi = fitted_values(y, pars)

    res = y[len(pars)-1:] - yi
    
    mean = np.mean(y)

    # Add mean to the residuals
    for i in range(len(res)):
        res.iloc[i] += mean

    return res

# This method selects the best AR model with a specified maximum order
# The best model is selected based on AIC value
def autoreg(data: pd.Series, max_lag=5):
    """ autoregressive (AR) models and perform functions related to AR modeling.
    
    Extended Summary
    ----------------
    Contains methods that fit series to autoregressive models and perform functions 
    related to AR modeling.
  
    NOTE: This function only accepts pandas series and dataframes as parameters. 

    Parameters
    ----------
    data : str
           a data file (.CSV or .RWL) or a pandas dataframe imported from dpl.readers().
    series : str
             an individual series within a chronology `data` file.
    lag : int, default 5
          nuber of years.
            
    Returns
    -------
    data :  pandas dataframe
        
    Examples
    --------
    >>> import dplpy as dpl 
    >>> data = dpl.readers("../tests/data/csv/file.csv")
    >>> dpl.autoreg(data['series name']) -> returns parameters of best fit AR model
                                            with maxlag of 5 (default) or other 
                                            specified number
    References
    ----------
    .. [1] https:/opendendro.org/dplpy-man/#autoreg
    
    """
    # validate data?
    if not isinstance(data, pd.Series):
        raise TypeError("Data argument should be pandas series. Received " + str(type(data)) + " instead.")

    
    max_allowable_lag = len(data.dropna())//2 - 1
    max_lag_used = max_lag if max_lag <= max_allowable_lag else max_allowable_lag

    # Need to change this to only ignore specific warnings instead of all
    with warnings.catch_warnings():
        warnings.filterwarnings("ignore")
        ar_data = ar_select_order(data.dropna(), max_lag_used, ic='aic', old_names=False)
    results = ar_data.model.fit()
    return results.params

# This function calculates the in-sample predicted values of a series,
# given an array containing the original data and the parameters for
# the AR model
def fitted_values(data_series, params):
    results = []
    
    for i in range((len(params)-1), len(data_series)):
        pred = params.iloc[0]
        for j in range(1, len(params)):
            pred += (params.iloc[j] * data_series.iloc[i-j])
        results.append(pred)
    return np.asarray(results)


__copyright__ = """
   dplPy for tree ring width time series analyses
   Copyright (C) 2024  OpenDendro

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""
__license__ = "GNU GPLv3"

#!/usr/bin/python
# -*- coding: utf-8 -*-

# Date: 5/12/2023
# Author: Ifeoluwa Ale
# Title: xdate.py
# Project: OpenDendro dplPy
# Description: Crossdating function for dplPy datasets.
#
# example usage from Python Console: 
# >>> import dplpy as dpl 
# >>> data = dpl.readers("../tests/data/csv/file.csv")
# >>> dpl.xdate(data)
# >>> dpl.xdate(data, prewhiten=False, corr="Pearson", show_flags=False)
# >>> dpl.xdate(data, slide_period=50, bin_floor=10, p_val=0.02)

from detrend import detrend
from autoreg import ar_func_series
from chron import chron
from stats import stats
import matplotlib.pyplot as plt
import pandas as pd
import numpy as np
import scipy
import re


# Main crossdating function, returns a dataframe of each series' segment correlations compared to the same
# segments in the master chronology.
def xdate(data: pd.DataFrame, prewhiten=True, corr="Spearman", slide_period=50, bin_floor=100, p_val=0.05, show_flags=True):
    """Crossdating function
    
    Extended Summary
    ----------------

    Parameters
    ----------
    data : pandas dataframe
           a data file (.CSV or .RWL), or an array imported from dpl.readers()
    prewhiten : boolean, default True
                prewhiten series using AR modeling
    corr : str, default Spearman
           correlation type, options: 'Pearson' or 'Spearman'
    slide_period : int, default 50
                   period window (years)
    bin_floor : int, default 100
                bin size
    p_val : float, default 0.05
            p-value, options: '0.05', '0.01', '0.001'
    show_flags : boolean, default True
                 show flags in the output
                   
    Returns
    -------
    data : pandas dataframe
    plot : matplotlib.pyplot figure
    
    Examples
    --------
    >>> ca533_rwi = dpl.detrend(ca533, fit="spline", method="residual", plot=False)
    # Crossdating of detrended data
    >>> dpl.xdate(ca533_rwi, prewhiten=True, corr="Spearman", slide_period=50, bin_floor=100, p_val=0.05, show_flags=True)

    References
    ----------
    .. [1] https:/opendendro.org/dplpy-man/#xdate
            
    """
    # Check types of inputs
    if not isinstance(data, pd.DataFrame):
        errorMsg = "Expected dataframe input, got " + str(type(data)) + " instead."
        raise TypeError(errorMsg)
    
    # Identify first and last valid indexes, for separating into bins.
    bins, bin_data = get_bins(data.first_valid_index(), data.last_valid_index(), bin_floor, slide_period)

    rwi_data = detrend(data, fit="horizontal", plot=False)

    # if detrending returns error, raise to output
    if isinstance(rwi_data, ValueError) or isinstance(rwi_data, TypeError):
        raise rwi_data

    # drop nans, prewhiten series if necessary
    df_start = pd.DataFrame(index=pd.Index(data.index))
    to_concat = [df_start]

    for series in rwi_data:
        nullremoved_data = rwi_data[series].dropna()
        if prewhiten is True:
            try:
                res = ar_func_series(nullremoved_data, get_ar_lag(nullremoved_data))
                offset = len(nullremoved_data) - len(res)
                to_concat.append(pd.Series(data=res, name=series, index=nullremoved_data.index.to_numpy()[offset:]))
            except ZeroDivisionError:
                print("Zero division error for series:", series, ". Dropping series.")
        else:
            to_concat.append(nullremoved_data)

    ready_series = pd.concat(to_concat, axis=1)

    ready_series_copy = ready_series.copy()
    ready_series = ready_series.rename_axis(data.index.name)

    series_names = []
    series_corr = []

    for series in sorted(ready_series):
        removed = ready_series_copy.pop(series)

        # create chronology with current series removed
        new_chron = chron(ready_series_copy, plot=False)["Mean RWI"]

        # correlate current series against chronology of remaining series
        inp = pd.concat([removed, new_chron], axis=1, join='inner')
        series_names.append(series)
        series_corr.append(correlate(inp, corr))

        flags = {"A":[], "B":[]}
        
        # evaluation of current series vs chronology of others by segments of years (the bins created earlier)
        for range in bins:
            # print(range) # useful for debugging but not necessary once operational
            start = int(re.split("(?<=\\d)-", range)[0])
            end = int(re.split("(?<=\\d)-", range)[1])
            if start >= removed.first_valid_index() and end <= removed.last_valid_index():
                segment = removed.loc[start:end]

                seg_corr, flag, flag_data = compare_segment(segment, new_chron, slide_period, corr, p_val, slide=show_flags)

                if flag is not None:
                    flags[flag].append(flag_data)
                bin_data[range].append(seg_corr)
            else:
                bin_data[range].append(np.nan)
        #ready_series_copy[series] = removed
        ready_series_copy = pd.concat([ready_series_copy, removed], axis=1)
        
        if show_flags is True:
            print_flags(series, flags)
    print()

    bin_res = pd.DataFrame.from_dict(bin_data)
    bin_res.set_index(pd.Index(series_names), inplace=True)

    return bin_res.transpose()

# Variation of xdate function that plots a graph that color codes the segment correlations. 
# Will be merged into original xdate function when completed, so that users can choose to
# show the graph by passing an optional argument.
def xdate_plot(data: pd.DataFrame, slide_period=50, bin_floor=10):
    plot_data = xdate(data, slide_period=slide_period, bin_floor=bin_floor, show_flags=False)
    bins = plot_data.index.to_numpy()

    # obtain a list of series names sorted by the start date
    data_stats = stats(data)
    series_by_start_date = data_stats.sort_values(by='first')['series']

    # Change the style of plot
    plt.style.use('seaborn-v0_8-darkgrid')

    years = data.index.to_numpy()

    # set width and height of the window based on the data
    dimensions = (max((years[-1] - years[0])//80, 8), max(len(data.columns)//2, 8))
    lin_wid = 7.5

    plt.figure(figsize=(dimensions))
 
    # separate plots for each series using the offset
    offset = (data.mean().mean() * 2)

    y_divisions = [] # needed to put series names on the y-axis
    num=0
    for column_name in series_by_start_date:
        num+=1
        first = np.nan
        second = np.nan
        penult = np.nan
        last = np.nan
        
        for i in range(0, len(bins)-1, 2):
            bin_range1 = list(map(int, bins[i].split("-")))
            bin_range2 = list(map(int, bins[i+1].split("-")))

            # plot for bin range 1 if valid
            if (bin_range1[0] >= data[column_name].first_valid_index() and bin_range1[1] <= data[column_name].last_valid_index()):
                range_1_color = get_graph_color(plot_data.loc[bins[i]][column_name])
                plt.plot(bin_range1, np.zeros((2,), dtype=int) + (offset * (num-1) + (offset/2)), marker='_', linewidth=lin_wid, alpha=0.9, color=range_1_color)
                first = np.nanmin([first, bin_range1[0]])
                second = np.nanmin([second, bin_range1[1]])

            # plot for bin range 2 if valid
            if (bin_range2[0] >= data[column_name].first_valid_index() and bin_range2[1] <= data[column_name].last_valid_index()):
                range_2_color = get_graph_color(plot_data.loc[bins[i+1]][column_name])
                plt.plot(bin_range2, np.zeros((2,), dtype=int) + (offset * (num-1)), marker='_', linewidth=lin_wid, alpha=0.9, color=range_2_color)
                penult = np.nanmax([penult, bin_range2[0]])
                last = np.nanmax([last, bin_range2[1]])
        
        # pad before and after first 2 and last 2 bins with greens
        pad_start_and_end_of_series_graph(data[column_name], first, second, penult, last, num, offset, lin_wid)
        y_divisions.append(offset*(num-1))

    # set y-axis to display series names at equal intervals, and x-axis to display years
    plt.yticks(y_divisions, series_by_start_date)
    plt.xlabel("Year")

    # Show the graph
    plt.show()

def pad_start_and_end_of_series_graph(series, first, second, penult, last, num, offset, lin_wid):
    if not np.isnan(first):
        first_range = [series.first_valid_index(), first]
        plt.plot(first_range, np.zeros((2,), dtype=int) + (offset * (num-1) + (offset/2)), marker='_', linewidth=lin_wid, alpha=0.9, color='#00ff00')

                       
    if not np.isnan(second):
        second_range = [series.first_valid_index(), second]
        plt.plot(second_range, np.zeros((2,), dtype=int) + (offset * (num-1)), marker='_', linewidth=lin_wid, alpha=0.9, color='#00ff00')
    
    if not np.isnan(penult):
        penult_range = [penult, series.last_valid_index()]
        plt.plot(penult_range, np.zeros((2,), dtype=int) + (offset * (num-1) + (offset/2)), marker='_', linewidth=lin_wid, alpha=0.9, color='#00ff00')

    if not np.isnan(last):
        last_range = [last, series.last_valid_index()]
        plt.plot(last_range, np.zeros((2,), dtype=int) + (offset * (num-1) + (offset/2)), marker='_', linewidth=lin_wid, alpha=0.9, color='#00ff00')


# Helper function that determines the color of a segment of the graph depending on the correlation value.
def get_graph_color(corr_val):
    if corr_val == np.nan:
        return '#00ff00'
    elif corr_val < 0.1:
        return '#ff0d1a'
    elif corr_val < 0.3:
        return '#add8ff'
    elif corr_val < 0.4:
        return '#9cc7ff'
    elif corr_val < 0.5:
        return '#33b6ff'
    elif corr_val < 0.6:
        return '#0033ff'
    elif corr_val < 0.7:
        return '#0000ff'
    elif corr_val < 0.8:
        return '#0000dd'
    elif corr_val < 0.9:
        return '#0000b3'
    elif corr_val < 1:
        return '#000099'

# Determines the max lag to use for AR modeling function.
def get_ar_lag(data):
    n = len(data)
    res = min(int(n-1), int(10 * np.log10(n)))
    return res

# Generates the bins given the first and last years of recorded data, the bin floor and
# the desired number of years in a period (bin).
def get_bins(first_year, last_year, bin_floor, slide_period):
    overlap = int(slide_period/2)
    if bin_floor == 0 or first_year % bin_floor == 0:
        start = first_year
    else:
        start = (int(first_year/bin_floor) * bin_floor) + bin_floor
    
    i = start

    bins = []
    bin_data = {}
    while i+slide_period-1 < last_year:
        period = str(i) + "-" + str(i+slide_period-1)
        bins.append(period)
        bin_data[period] = []
        i += overlap
    return bins, bin_data

# Returns the correlation value of the given data. Can find Spearman or Pearson's
# correlations
def correlate(data, type):
    if type == "Spearman":
        return scipy.stats.spearmanr(data, axis=0).correlation
    elif type == "Pearson":
        return np.corrcoef(data, rowvar=False)[0, 1]

def get_crit(alpha=0.01, n=50, type="one-tailed"):
    if type == "two-tailed":
        tcrit = scipy.stats.t.ppf(1-(alpha/2),n-2)
    else:
        tcrit = scipy.stats.t.ppf(1-(alpha),n-2)
    
    cc = pow((tcrit/np.sqrt(n-2)), 2)
    rcrit = np.sqrt(cc/(1+cc))
    return rcrit
  
# Compares segments to the mean values of the master chronology excluding the current series.
# This is where flags in dating are detected.
def compare_segment(segment, new_chron, slide_period, correlation_type, p_val, slide=True, left_most=-10, right_most=10):
    flag = None

    if segment.size < slide_period:
        return
    series_name = segment.name
    data = pd.concat([segment, new_chron], axis=1, join='inner')
    original = correlate(data, correlation_type)

    # Will set threshold to 99% confidence p value that is based on segment length.
    if original < get_crit(p_val):
        flag = "A"

    if not slide:
        return original, flag, None
    
    best_lag = 0
    best_coeff = original
    segment_data = []

    for shift in range(left_most, right_most+1):
        shifted = data[series_name].copy(deep=False)
        shifted.index += shift
        overlapping_df = pd.concat([shifted, new_chron], axis=1, join='inner').dropna()
        if overlapping_df.size == slide_period * 2:
            new_coeff = correlate(overlapping_df, correlation_type)
            segment_data.append(('{0:.2f}'.format(new_coeff)).rjust(5))
            if new_coeff > best_coeff:
                # for debugging print(segment.name, segment.first_valid_index(), shift)
                best_lag = shift
                best_coeff = new_coeff
        else:
            segment_data.append("     ")
    
    if (best_lag != 0 and abs(best_coeff - original) >= 0.08):
        flag = "B"

    if flag is not None:    
    # return original, flag ("A", "B" or None), with segment data, (None if flag is None and an array if otherwise.)
        segment_data = [(str(segment.first_valid_index()) + "-" + str(segment.last_valid_index())).rjust(12), str(best_lag).rjust(4)] + segment_data
        return original, flag, segment_data
    else:
        return original, flag, None

# Prints the flags that had been detected for a series.
def print_flags(series_name, flags):
    if flags["A"] == [] and flags["B"] == []:
        return

    print("Flags for", series_name)
    if flags["A"] != []:
        print("[A] Segment  High   -10    -9    -8    -7    -6    -5    -4    -3    -2    -1     0    +1    +2    +3    +4    +5    +6    +7    +8    +9   +10")
        for flag in flags["A"]:
            print(" ".join(flag))
    if flags["B"] != []:
        print("[B] Segment  High   -10    -9    -8    -7    -6    -5    -4    -3    -2    -1     0    +1    +2    +3    +4    +5    +6    +7    +8    +9   +10")
        for flag in flags["B"]:
            print(" ".join(flag))
    
    print()



__copyright__ = """
   dplPy for tree ring width time series analyses
   Copyright (C) 2024  OpenDendro

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

__license__ = "GNU GPLv3"

#!/usr/bin/python
# -*- coding: utf-8 -*-

# Date: 11/1/2022
# Author: Ifeoluwa Ale
# Title: detrend.py
# Description: Detrends a given series or data frame, first by fitting data to curve(s), 
#              with spline(s) as the default, and then by calculating residuals or differences 
#              compared to the original data (residuals by default).

import pandas as pd
import matplotlib.pyplot as plt
from smoothingspline import spline
import curvefit

def detrend(data: pd.DataFrame | pd.Series, fit="spline", method="residual", plot=True, period=None):
    if isinstance(data, pd.DataFrame):
        res = pd.DataFrame(index=pd.Index(data.index))
        to_add = [res]
        for column in data.columns:
            to_add.append(detrend_series(data[column], fit, method, plot, period))
        output_df = pd.concat(to_add, axis=1)
        return output_df.rename_axis(data.index.name)
    
    elif isinstance(data, pd.Series):
        return detrend_series(data, fit, method, plot, period)
    else:
        raise TypeError("argument should be either pandas dataframe or pandas series.")

# Takes a series as input and by default fits it to a spline, then 
# detrends it by calculating residuals
# Can specify what type of alternate curve-fits, or if the user
# would like to detrend by using differences
# Need to add series names to the top of the plots, and display the plots side by side
def detrend_series(data: pd.Series, fit, method, plot, period=None):
    """Detrends a given series or dataframe
    
    Extended Summary
    ----------------
    Detrends a given series or dataframe, first by fitting data to curve(s), 
    with 'spline' as the default, and then by calculating residuals 
    (default = 'residual') or differences ('difference') compared to the original data. 
    Other supported curve fitting methods are 'ModNegex' (modified negative exponential), 
    'Hugershoff', 'linear', 'horizontal'.
                  
    Parameters
    ----------
    data: pandas dataframe
          a data frame loaded using dpl.readers()
    series_name: str
                 name of the series to be detrended
    fit: str, default spline 
         fitting method of curves, e.g., 'horizontal', 'Hugershoff', 'linear', 'ModNegex' (modified negative exponential), and 'spline'.
    method : str, default residual
             intercomparison method, options: 'difference' or 'residual'.
    plot : boolean, default True
           plot the results.
    
    Returns
    -------
    data: pandas dataframe   
    
    Examples
    --------
    >>> import dplpy as dpl
    >>> data = dpl.readers("../tests/data/csv/file.csv")
    >>> dpl.detrend(data)
    # Detrending a series part of the dataframe
    >>> dpl.detrend(data["<series>"])
    # Detrending function and its options
    >>> dpl.detrend(data["<series>"], fit="<fitting method>", method="<comparison method>", plot=<True/False>)    
    
    References
    ----------
    .. [1] https:/opendendro.org/dplpy-man/#detrend
         
    """
    series_name = data.name
    nullremoved_data = data.dropna()
    x = nullremoved_data.index.to_numpy()
    y = nullremoved_data.to_numpy()

    if fit == "spline":
        yi = spline(x, y, period)
    elif fit == "ModNegEx":
        yi = curvefit.negex(x, y)
    elif fit == "Hugershoff":
        yi = curvefit.hugershoff(x, y)
    elif fit == "linear":
        yi = curvefit.linear(x, y)
    elif fit == "horizontal":
        yi = curvefit.horizontal(x, y)
    else:
        # give error message for unsupported curve fit
        raise ValueError("unsupported keyword for curve-fit type. See documentation for more info.")
    
    if method == "residual":
        detrended_data = residual(y, yi)
    elif method == "difference":
        detrended_data = difference(y, yi)
    else:
        # give error message for unsupported detrending method
        raise ValueError("unsupported keyword for detrending method. See documentation for more info.")
    
    if plot:
        fig, axes = plt.subplots(nrows=1, ncols=2, figsize=(7,3))
    
        axes[0].plot(x, y, "k-", x, yi, "r-", linewidth=2)
        axes[0].set_xlabel('Year')
        axes[0].set_ylabel('Ring Width')
        axes[0].set_title(series_name + " curve fit to " + fit)

        axes[1].plot(x, detrended_data, 'k-')
        axes[1].set_xlabel('Year')
        axes[1].set_ylabel('Index')
        axes[1].set_title(series_name + " detrended by " + method)

        fig.tight_layout()
    
        plt.show()

    return pd.Series(detrended_data, index=pd.Index(data=x, name="Year"), name=series_name).combine(data, pick_first)

def pick_first(a, b):
    return a

# Detrends by finding ratio of original series data to curve data
def residual(y, yi):
    return y/yi


# Detrends by finding difference between original series data and curve fit data
def difference(y, yi):
    return y - yi


from rbar import get_running_rbar, mean_series_intercorrelation
from chron import chron
import numpy as np
import pandas as pd
import warnings


def chron_stabilized(rwi_data: pd.DataFrame, win_length=50, min_seg_ratio=0.33, biweight=True, running_rbar=False):
    if not isinstance(rwi_data, pd.DataFrame):
        raise TypeError("Expected data input to be a pandas dataframe, not " + str(type(rwi_data)) + ".")
    
    
    num_years = rwi_data.shape[0]

    if win_length > num_years:
        raise ValueError("Window length should not be greater than the number of rows in the dataset")
    
    if min_seg_ratio <= 0 or min_seg_ratio > 1:
        raise ValueError("min_seg_ratio cannot be <= 0 or > 1")
    
    if win_length < 0.3*num_years or win_length >= 0.5*num_years:
        warnings.warn("We recommend using a window length greater than 30%% but less than 50%% of the chronology length\n")
    
    print("Generating variance stabilized chronology...\n")

    # give rbar function a range of years (window length) to calculate rbar for
    # calculate rbar for that window, using either osborn's or frank's or 67spline
    # get rbar for each relevant segment of the dataframe


    mean_val = rwi_data.mean().mean()

    zero_mean_data = rwi_data - mean_val

    rbar_array = np.zeros(zero_mean_data.shape[0])

    if win_length % 2 == 0:
        target = (win_length)/2
    else:
        target = (win_length-1)/2
    
    for i in range(num_years-win_length + 1):
        data_segment = zero_mean_data[i:i + win_length]
        if data_segment.shape[0] < win_length:
            continue
        target_index = int(i + target)
        rbar_array[target_index] = get_running_rbar(data_segment, min_seg_ratio)

    rbar_array = pad_rbar_array(rbar_array)

    reg_chron = chron(zero_mean_data, biweight=biweight, plot=False)

    mean_rwis = reg_chron["Mean RWI"].to_numpy()
    samp_deps = reg_chron["Sample depth"].to_numpy()
    denom = np.multiply(samp_deps-1, rbar_array) + 1

    n_eff = np.minimum(np.divide(samp_deps, denom), samp_deps)
    rbar_const = mean_series_intercorrelation(zero_mean_data, "pearson", min_seg_ratio)
    stabilized_means = np.multiply(mean_rwis, np.sqrt(n_eff * rbar_const))

    if running_rbar:
        stabilized_chron =  pd.DataFrame(data={"Adjusted CRN": stabilized_means + mean_val, "Running rbar": rbar_array, "Sample depth": samp_deps}, index=reg_chron.index)
    else:
        stabilized_chron =  pd.DataFrame(data={"Adjusted CRN": stabilized_means + mean_val, "Sample depth": samp_deps}, index=reg_chron.index)

    print("SUCCESS!\n")
    return stabilized_chron
    
def pad_rbar_array(rbar_array):
    # double check that rbar cannot be 0
    first = 0
    first_valid = 0
    for val in rbar_array:
        if val != 0 and not np.isnan(val):
            first = val
            break
        first_valid += 1

    last = 0
    last_valid = len(rbar_array) - 1
    for val in np.flip(rbar_array):
        if val != 0 and not np.isnan(val):
            last = val
            break
        last_valid -= 1
    
    rbar_array[:first_valid] = np.full(first_valid, first) #should be  np.full(first_valid, 1)
    rbar_array[last_valid:] = np.full(len(rbar_array) - last_valid, last) #should be np.full(len(rbar_array) - last_valid, last)

    return rbar_array


__copyright__ = """
   dplPy for tree ring width time series analyses
   Copyright (C) 2024  OpenDendro

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

__license__ = "GNU GPLv3"

#!/usr/bin/python
# -*- coding: utf-8 -*-

# Date: 11/1/2022
# Author: Ifeoluwa Ale
# Title: curvefit.py
# Description: This file contains helper functions which fit data
#              from a series to curves. The curves included are hugershoff,
#              modified negative exponential, linear and horizontal.

import numpy as np
from scipy.optimize import curve_fit

# Modified hugershoff function
def hugershoff_function(x, a, b, c, d):
    return a*(x**b)*np.exp(c*x) + d

# Attempt to fit a hugershoff curve to the series
def hugershoff(x, y):
    xi = np.arange(1, len(y)+1)
    pars, unk= curve_fit(hugershoff_function, xi, y, bounds=([0, -2, -np.inf, min(y)], [np.inf, 2, 0, max(y)]), 
                            p0=[max(y)-min(y), 0, 0, y[0]])
    a, b, c, d = pars

    yi = hugershoff_function(xi, a, b, c, d)
    return yi

# Modified negative exponential function
def negex_function(x, a, b, k):
    return a * np.exp(b * x) + k

# Attempt to fit a negative exponential curve to the series
def negex(x, y):
    xi = np.arange(1, len(y)+1)
    pars, unk= curve_fit(negex_function, xi, y, bounds=([0, -np.inf, 0], [np.inf, 0, np.inf]))
    a, b, k = pars

    yi = negex_function(xi, a, b, k)
    return yi

# Fit a horizontal line to the series
def horizontal(x, y):
    yi = np.asarray([np.mean(y)] * len(x))
    return yi
    
# Equation of a straight line
def line_function(x, m, c):
    return (m * x) + c

# Fit a line to the series
def linear(x, y, bounds=False):
    if bounds is False:
        pars, unk = curve_fit(line_function, x, y)
    else:
        pars, unk = curve_fit(line_function, x, y, bounds=([-np.inf, -np.inf], [0, np.inf]))
    m, c = pars
    yi = line_function(x, m, c)
    return yi


from __future__ import print_function

__copyright__ = """
   dplPy for tree ring width time series analyses
   Copyright (C) 2024  OpenDendro

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""
__license__ = "GNU GPLv3"

#!/usr/bin/python
# -*- coding: utf-8 -*-

# Date: 11/1/2022
# Author: Ifeoluwa Ale
# Title: chron.py
# Project: OpenDendro dplPy
# Description: Creates a mean value chronology for a dataset, typically the ring width indices of a detrended series.
#              Takes three optional arguments 'biweight', 'prewhiten', and 'plot'. They determine whether to find means using tukey's
#              biweight robust mean (default True), whether to prewhiten data by fitting to an AR model (default False), and
#              whether to plot the results of the chronology (default True).


import matplotlib.pyplot as plt
import pandas as pd
import numpy as np
from tbrm import tbrm
from autoreg import ar_func

# Main function for creating chronology of series. Formats input, prewhitens if necessary
# and produces output mean value chronology in a dataframe.
def chron(rwi_data: pd.DataFrame, biweight=True, prewhiten=False, plot=True):
    """Creates a mean value chronology for a dataset
    
    Extended Summary
    ----------------
    Creates a mean value chronology for a dataset, typically the ring width 
    indices of a detrended series. Takes three optional arguments 'biweight', 
    'prewhiten', and 'plot'. They determine whether to find means using Tukey's 
    bi-weight robust mean (default 'True'), whether to prewhiten data by fitting 
    to an AR model (default 'False'), and whether to plot the results of the 
    chronology (default 'True').
    
    Parameters
    ----------
    data : pandas dataframe
           a pandas dataframe imported from dpl.readers()
    biweight : boolean, default True
               use Tukey's bi-weight robust mean   
    prewhiten : boolean, default False   
                run pre-whitening on the time series
    plot : boolean, default True 
           plot the results    
        
    Returns
    -------
    data: pandas dataframe
        a data frame loaded using readers
    
    Examples
    --------
    >>> import dplpy as dpl 
    >>> data = dpl.readers("../tests/data/csv/file.csv")
    >>> rwi_data = dpl.detrend(data)
    >>> dpl.chron(rwi_data)
    >>> dpl.chron(rwi_data, prewhiten=True)
    >>> chron_data = dpl.chron(rwi_data, biweight=False, plot=False)
    
    References
    ----------
    .. [1] https:/opendendro.org/dplpy-man/#chron
    
    """
    if not isinstance(rwi_data, pd.DataFrame):
        raise TypeError("Expected pandas dataframe as input, got " + str(type(rwi_data)) + " instead")
    
    chron_data = {}
    for series in rwi_data:
        series_data = rwi_data[series].dropna()
        for year in series_data.keys().tolist():
            if year not in chron_data:
                chron_data[year] = [1, series_data[year]]
            else:
                chron_data[year][0] += 1
                chron_data[year].append(series_data[year])
    
    years, means, depths = get_chron_info(chron_data, biweight)
    chron_res = pd.DataFrame(data={"Year":years})
    chron_res = pd.concat([chron_res, pd.Series(data=means, name="Mean RWI")], axis=1)
    
    if prewhiten:
        whitened_means = get_whitened_chron_info(rwi_data, chron_data, biweight)
        chron_res = pd.concat([chron_res, pd.Series(data=whitened_means, name="Mean Res")], axis=1)
    else:
        whitened_means = None
    
    chron_res = pd.concat([chron_res, pd.Series(data=depths, name="Sample depth")], axis=1)
    chron_res.set_index('Year', inplace = True, drop = True)

    if plot:
        plot_chron(years, depths, means, whitened_means)
    
    return chron_res

# Does the work of creating the actual chronology by finding the mean RWI for each year
# in the dataset and keeping track of the sample size.
def get_chron_info(chron_data, biweight):
    years = []
    means = []
    depths = []
    
    for year in sorted(chron_data):
        years.append(year)
        if biweight:
            means.append(tbrm(chron_data[year][1:]))
        else:
            means.append(sum(chron_data[year][1:])/chron_data[year][0])
        depths.append(chron_data[year][0])

    return years, means, depths

# Does the work of creating a chronology when the data has to be fit to an AR model
# first (i.e. prewhitened).
def get_whitened_chron_info(rwi_data, chron_data, biweight):
    whitened_data = {}
    ar_fit_data = {}

    for series in rwi_data:
        series_data = rwi_data[series].dropna()
        series_years = series_data.keys().tolist()
        ar_fit_data[series] = ar_func(series_data)

        offset = len(series_years) - len(ar_fit_data[series])
        i = 0
        for year in series_years[offset:]:
            if year not in whitened_data:
                whitened_data[year] = [1, ar_fit_data[series].iloc[i]]
            else:
                whitened_data[year][0] += 1
                whitened_data[year].append(ar_fit_data[series].iloc[i])
            i += 1

    whitened_means = []
    for year in sorted(chron_data):
        if year not in whitened_data:
            whitened_means.append(np.nan)
        elif biweight:
            whitened_means.append(tbrm(whitened_data[year][1:]))
        else:
            whitened_means.append(sum(whitened_data[year][1:])/whitened_data[year][0])
    return whitened_means

# Plots the data created by the chronology
def plot_chron(years, depths, means, whitened_means):
    # create figure and axis objects with subplots()
    fig,ax = plt.subplots()

    if whitened_means is not None:
        y_val = whitened_means
        y_label = "Mean Res"
    else:
        y_val = means
        y_label = "Mean RWI"

    # make plot of RWI means
    ax.plot(years, y_val, "k-")
    ax.set_xlabel("Year", fontsize = 14)
    ax.set_ylabel(y_label, fontsize=14)

    # twin object for two different y-axis on the sample plot
    ax2=ax.twinx()
    # make plot of sample depths
    ax2.fill_between(years, depths, color=((0.2, 0.6, 0.9, 0.3)))
    ax2.set_ylabel("Sample depth",fontsize=14)
    fig.set_size_inches(14, 8)
    plt.show()


from __future__ import print_function

__copyright__ = """
   dplPy for tree ring width time series analyses
   Copyright (C) 2024  OpenDendro

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""
__license__ = "GNU GPLv3"

#!/usr/bin/python
# -*- coding: utf-8 -*-

# Date: 5/27/2022
# Author: Ifeoluwa Ale
# Title: summary.py
# Description: Generates a summary of each series recorded in Tucson format and CSV format files
# example usage:
# >>> import dplpy as dpl 
# >>> data = dpl.readers("../tests/data/csv/file.csv")
# >>> dpl.summary(data)
#
# >>> dpl.summary("../tests/data/csv/file.csv")
# >>> Note: for file pathname inputs, only CSV and RWL file formats are accepted

# Create Summaries for Tucson (*rwl) files
import pandas as pd

from readers import readers
def summary(inp: pd.DataFrame | str):
    """ Summarizes a chronology
    
    Extended Summary
    ----------------
    This function summarizes a chronology from an array
    
    Parameters
    ----------
    data : str
           a data file (.CSV or .RWL) or a pandas dataframe imported from dpl.readers().
   
    Returns
    -------
    data : pandas dataframe
    
    Examples
    --------
    >>> import dplpy as dpl
    >>> data = dpl.readers("../tests/data/csv/file.csv")
    >>> dpl.summary(data)
    
    References
    ----------
    .. [1] https:/opendendro.org/dplpy-man/#summary
    
    """
    if isinstance(inp, pd.DataFrame):
        series_data = inp
    elif isinstance(inp, str):
        series_data = readers(inp)
    else:
        errorMsg = """
Unable to generate summary report. Input must be string path to file to be read
or Dataframe object.

Note: for file pathname inputs, only CSV and RWL file formats are accepted

Example usages:

>>> import dplpy as dpl
>>> data = dpl.readers('../tests/data/csv/file.csv')
>>> dpl.summary(data)
>>> dpl.summary('../tests/data/csv/file.csv')

"""
        raise TypeError(errorMsg)

    summary = series_data.describe()
    return summary


__copyright__ = """
   dplPy for tree ring width time series analyses
   Copyright (C) 2024  OpenDendro

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""
__license__ = "GNU GPLv3"

#!/usr/bin/python
# -*- coding: utf-8 -*-

# Date: 5/12/2023
# Author: Ifeoluwa Ale
# Title: series_corr.py
# Project: OpenDendro dplPy
# Description: Crossdating function that focuses on the comparison of one series to the
# master chronology.
#
# example usage from Python Console: 
# >>> import dplpy as dpl 
# >>> data = dpl.readers("../tests/data/csv/file.csv")
# >>> dpl.series_corr(data, "series_name")
# >>> dpl.series_corr(data, "series_name", prewhiten=False, corr="Pearson", bin_floor=10)

from detrend import detrend
from autoreg import ar_func_series
from chron import chron
from xdate import get_ar_lag, correlate, compare_segment, get_crit

import pandas as pd
import numpy as np
import matplotlib.pyplot as plt

# Analyzes the crossdating of one series compared to the master chronology
def series_corr(data: pd.DataFrame, series_name: str, prewhiten=True, corr="Spearman", seg_length=50, bin_floor=100, p_val=0.05):
    """Crossdating correlation function
    
    Extended Summary
    ----------------
    Crossdating correlation function that focuses on the comparison of one series to the master chronology.
    
    Parameters
    ----------
    data : pandas dataframe
           a data file (.CSV or .RWL), or an array imported from dpl.readers()
    series : str    
             a series name from the dataframe
    prewhiten : boolean, default False
                run pre-whitening on the time series, options: 'True' or 'False'.
    corr : str, default 'Spearman'
           select correlation type if 'prewhiten=True', options: 'Pearson' or 'Spearman'.
    seg_length :  int, default 50
                  segment length (years).
    bin_floor : int, default 100
                select bin size.
    p_val : double, default 0.05
            select a p-value, e.g., '0.05', '0.01', '0.001'.
    plot :  boolean, default True
            plot the output.
    
    Returns
    -------
    data : pandas dataframe
    
    Examples
    --------
    >>> dpl.series_corr(ca533, "CAM191", prewhiten=False, corr="Pearson", bin_floor=10)    
    
    References
    ----------
    .. [1] https:/opendendro.org/dplpy-man/#series_corr
    
    """
    # Check types of inputs
    if not isinstance(data, pd.DataFrame):
        errorMsg = "Expected dataframe input, got " + str(type(data)) + " instead."
        raise TypeError(errorMsg)
    
    if not isinstance(series_name, str):
        errorMsg = "Expected string input as series name, got " + str(type(series_name)) + " instead."
        raise TypeError(errorMsg)

    if series_name not in data.columns:
        errorMsg = "Series named " + series_name + " not found in provided dataframe."
        raise ValueError(errorMsg)

    rwi_data = detrend(data, fit="horizontal", plot=False)
 
    # drop nans, prewhiten series if necessary
    df_start = pd.DataFrame(index=pd.Index(data.index))
    to_concat = [df_start]
    for series in rwi_data:
        nullremoved_data = rwi_data[series].dropna()
        if prewhiten is True:
            res = ar_func_series(nullremoved_data, get_ar_lag(nullremoved_data))
            offset = len(nullremoved_data) - len(res)
            to_concat.append(pd.Series(data=res, name=series, index=nullremoved_data.index.to_numpy()[offset:]))
        else:
            to_concat.append(nullremoved_data)
    ready_series = pd.concat(to_concat, axis=1)
    ready_series = ready_series.rename_axis(data.index.name)

    removed = ready_series.pop(series_name)
    new_chron = chron(ready_series, plot=False)["Mean RWI"]

    inp = pd.concat([removed, new_chron], axis=1, join='inner')
    correlate(inp, corr)

    data_first = data.first_valid_index()
    data_last = data.last_valid_index()
    ser_first = removed.first_valid_index()
    ser_last = removed.last_valid_index()

    start, end = get_rel_range(data_first, data_last, ser_first, ser_last, bin_floor, seg_length)
    
    plt.style.use('seaborn-v0_8-darkgrid')
    wid = max((end - start)//30, 1)
    hei = 10
    base_corr = get_crit(p_val)
    
    dimensions = (wid, hei)
    plt.figure(num=1, figsize=(dimensions))
    plt.grid(True)

    years = []
    corrs = []

    second_plot = []
    # Find correlations at segment whose year is at center. So segments should be from i-25 to i+25, where i starts from bin start
    for i in range(start, end):
        segment = removed.loc[i-(seg_length//2):i+(seg_length//2)-1]

        if segment.size != seg_length:
            continue
        seg_corr, flag, flag_data = compare_segment(segment, new_chron, seg_length, corr, p_val, slide=False)

        years.append(i)
        corrs.append(seg_corr)

        if (((i-start-(seg_length//2)) % seg_length == 0)  or (((i-start-(seg_length//2)) % seg_length) == seg_length//2)):
            seg_range = [i-seg_length//2, i+seg_length//2]
            seg_corr_y = [seg_corr, seg_corr]
            plt.plot(seg_range, seg_corr_y, color="k")
            second_plot.append(analyze_segment(segment, new_chron, seg_length, corr))

            
    plt.plot(years, corrs, color="k")
    plt.plot([years[0]-seg_length, years[-1]+seg_length], [base_corr, base_corr], linestyle="dashed", color="k")
    plt.xlim([years[0]-seg_length, years[-1]+seg_length])
    plt.xlabel("Year")
    plt.ylabel("Correlation")

    plt.figure(num=2)
    plt.style.use('_mpl-gallery')
    cols = 5
    rows = (len(second_plot) // 5) + 1
    fig, axes = plt.subplots(nrows=rows, ncols=cols, figsize=(14,14))

    j = 0
    for lags in second_plot:
        row = j // 5
        col = j % 5
        x_vals = np.arange(-5, 6, 1)
        y_vals = lags

        axes[row][col].stem(x_vals, y_vals)
        axes[row][col].plot([-5, 5], [base_corr, base_corr], linestyle="dashed")
        axes[row][col].plot([-5, 5], [-base_corr, -base_corr], linestyle="dashed")
        axes[row][col].set_xlabel('Lag')
        axes[row][col].set_ylabel('Correlation')
        axes[row][col].set_title(str(start + ((seg_length//2)*j)) + '.' + str(start + ((seg_length//2)*j) + seg_length - 1))
        axes[row][col].set_xlim([-6, 6])
        axes[row][col].set_ylim([-0.5, 1])
        j += 1
    
    while j < (rows*cols):
        row = j // 5
        col = j % 5
        axes[row][col].set_axis_off()
        j += 1

    fig.tight_layout()
    plt.show()

# Gets correlation data for segments lagged by 5 years forwards and backwards.
def analyze_segment(segment, new_chron, slide_period, correlation_type):
    if segment.size < slide_period:
        return
    series_name = segment.name
    data = pd.concat([segment, new_chron], axis=1, join='inner')
    
    segment_data = []

    for shift in range(-5, 6):
        shifted = data[series_name].copy(deep=False)
        shifted.index += shift
        overlapping_df = pd.concat([shifted, new_chron], axis=1, join='inner').dropna()
        if overlapping_df.size == slide_period * 2:
            new_coeff = correlate(overlapping_df, correlation_type)
            segment_data.append(new_coeff)
        else:
            segment_data.append(0)
    return segment_data

# Finds the effective first and last year of crossdating for the series depending on the
# values of bin floor and segment length.
def get_rel_range(data_first, data_last, series_first, series_last, bin_floor, seg_len):
    overlap = seg_len/2

    if bin_floor == 0 or data_first % bin_floor == 0:
        start = data_first
    else:
        start = (int(data_first/bin_floor) * bin_floor) + bin_floor

    rel_start = 9999
    rel_stop = 0
    i = start
    while i+seg_len-1 < data_last:
        if i >= series_first:
            rel_start = min(rel_start, i)
            if i <= series_last:
                rel_stop = max(rel_stop, i+seg_len-1)
            else:
                break
        i += overlap

    return int(rel_start), int(rel_stop)



__copyright__ = """
   dplPy for tree ring width time series analyses
   Copyright (C) 2024  OpenDendro

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

__license__ = "GNU GPLv3"

#!/usr/bin/python
# -*- coding: utf-8 -*-

# Date: 9/8/2022
# Author: Ifeoluwa Ale
# Title: plot.py
# Description: Generates plots of tree ring with data from dataframes. 
#              Currently capable of generating line (default), spag and seg plots.
#
# example usages: 
# >>> import dplpy as dpl 
# >>> data = dpl.readers("../tests/data/csv/file.csv")
# >>> data = dpl.plot(data)
# >>> data = dpl.plot(data[[SERIES_1, SERIES_2, SERIES_3]], type="seg")
# >>> data = dpl.plot("../tests/data/csv/file.csv")
# 

import pandas as pd
import matplotlib.pyplot as plt
from readers import readers
from stats import stats

def plot(inp: pd.DataFrame | str, type="seg"):
    """Plots a given dataframe
    
    Extended Description
    --------------------
    Plots a given dataframe or series of a specific dataframe in either 
    line ('default'), spaghetti ('spag') or segment ('seg') plots.
                 
    Parameters
    ----------
    data : str 
           a data frame loaded using dpl.readers()
    series : str
             a single time series within the data array
    type : str, default line
           type of plot to generate, e.g., line, spaghetti ('spag'), or segment ('seg') or 'line'.

    Returns
    -------
    plot : matplotlib.pyplot figure
           a plot of the data
    
    Examples
    --------
    >>> dpl.plot(<data>)
    # Plot series subset of dataframe with a specified plot type
    >>> dpl.plot(<data>["<series>"], type=<plot type>)
    
    References
    ----------
    .. [1] https:/opendendro.org/dplpy-man/#plot
    
    """
    if isinstance(inp, pd.DataFrame):
        series_data = inp
    elif isinstance(inp, str):
        series_data = readers(inp)
    else:
        return

    if type == "line":
        plt.plot(series_data)
    elif type == "spag":
        spag_plot(inp)
    elif type == "seg":
        seg_plot(inp)
    else:
        raise ValueError("Unsupported plot type.")

def spag_plot(data):
    # obtain a list of series names sorted by the start date
    data_stats = stats(data)
    series_by_start_date = data_stats.sort_values(by='first')['series']

    # Change the style of plot
    plt.style.use('seaborn-v0_8-darkgrid')

    years = data.index.to_numpy()

    # set width and height of the window based on the data
    dimensions = (max((years[-1] - years[0])//80, 1), max(len(data.columns)//3, 1))
    plt.figure(figsize=(dimensions))
 
    # separate plots for each series using the offset
    offset = (data.mean().mean() * 2)

    y_divisions = [] # needed to put series names on the y-axis
    num=0
    for column_name in series_by_start_date:
        num+=1
        plt.plot(years, data[column_name].to_numpy() + (offset * (num-1)), marker='', linewidth=1, alpha=0.9, color='k')
        y_divisions.append(offset*(num-1))

    # set y-axis to display series names at equal intervals, and x-axis to display years
    plt.yticks(y_divisions, series_by_start_date)
    plt.xlabel("Year")

    # Show the graph
    plt.show()

def seg_plot(data):
    # obtain a list of series names sorted by the start date
    data_stats = stats(data)
    series_by_start_date = data_stats.sort_values(by='first')['series']

    # Change the style of plot
    plt.style.use('seaborn-v0_8-darkgrid')

    years = data.index.to_numpy()

    # set width and height of the window based on the data
    dimensions = (max((years[-1] - years[0])//80, 1), max(len(data.columns)//3, 1))
    plt.figure(figsize=(dimensions))
 
    # separate plots for each series using the offset
    offset = (data.mean().mean() * 2)

    y_divisions = [] # needed to put series names on the y-axis
    num=0
    for column_name in series_by_start_date:
        num+=1
        plt.plot(years, (data[column_name].to_numpy() - data[column_name].to_numpy()) + (offset * (num-1)), marker='', linewidth=1, alpha=0.9, color='k')
        y_divisions.append(offset*(num-1))

    # set y-axis to display series names at equal intervals, and x-axis to display years
    plt.yticks(y_divisions, series_by_start_date)
    plt.xlabel("Year")

    # Show the graph
    plt.show()


from __future__ import print_function

__copyright__ = """
   dplPy for tree ring width time series analyses
   Copyright (C) 2024  OpenDendro

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

__license__ = "GNU GPLv3"

#!/usr/bin/python
# -*- coding: utf-8 -*-

# Date: 9/8/2022
# Author: Ifeoluwa Ale
# Project: OpenDendro- Readers
# Description: Reads data from supported file types (*.CSV and *.RWL)
#              and stores them in a dataframe

import os
import sys
import pandas as pd
import numpy as np

def readers(filename: str, skip_lines=0, header=False):
    """Imports a common ring width data file
    
    Extended Summary
    ----------------
    This function imports common ring width data files (.csv, .rwls) as arrays
    
    Parameters
    ----------
    filename : str
               a data file (.CSV or .RWL) or a pandas dataframe imported from dpl.readers().
    
    Returns
    -------
    data : pandas dataframe
    
    Examples
    --------
    >>> import dplpy as dpl
    >>> data = dpl.readers("../tests/data/csv/file.csv")
    >>> data = dpl.readers("../tests/data/csv/file.rwl", header=True)
    
    References
    ----------
    .. [1] https:/opendendro.org/dplpy-man/#readers

    """
    FORMAT = "." + filename.split(".")[-1]
    print("\nAttempting to read input file: " + os.path.basename(filename) + " as " + FORMAT + " format\n")
    
    # open the input file and read its data into a pandas dataframe
    if filename.upper().endswith(".CSV"):
        series_data = pd.read_csv(filename, skiprows=skip_lines)
    elif filename.upper().endswith(".RWL"):
        series_data = process_rwl_pandas(filename, skip_lines, header)
    elif filename.upper().endswith(".RAW"):
        series_data = process_rwl_pandas(filename, skip_lines, header)
    else:
        errorMsg = """

Unable to read file, please check that you're using a supported type
Accepted file types are .csv and .rwl

Example usages:
>>> import dplpy as dpl
>>> data = dpl.readers('../tests/data/csv/filename.csv')
>>> data = dpl.readers('../tests/data/rwl/filename.rwl'), header=True
"""
        
        raise ValueError(errorMsg)

    # If no data is returned, then an error was encountered when reading the file.
    if series_data is None:
        errorMsg = """
        Error reading file. Check that file exists and that file formatting is consistent with {format} format.
        If your file contains headers, run dpl.headers(file_path, header=True)
        """.format(format=FORMAT)
        raise ValueError(errorMsg)
    series_data.set_index('Year', inplace = True, drop = True)

    # Display message to show that reading was successful
    print("\nSUCCESS!\nFile read as:", FORMAT, "file\n")

    # Display names of all the series found
    print("Series names:")
    print(list(series_data.columns), "\n")
    return series_data

# Process data from .rwl file and store data in a pandas dataframe.
def process_rwl_pandas(filename, skip_lines, header):
    if header is True:
        skip_lines += 3 # working with the assumption that headers are 3 lines long

    with open(filename, "r") as rwl_file:
        file_lines = rwl_file.readlines()[skip_lines:]
    
    rwl_data, first_date, last_date = read_rwl(file_lines)
    if rwl_data is None:
        return None

    # create an array of indexes for the dataframe
    indexes = []
    for i in range(first_date, last_date):
        indexes.append(i)
    
    df = pd.DataFrame(data={"Year":indexes})

    # store raw data in pandas dataframe
    for series in rwl_data:
        series_data = []
        for i in range(first_date, last_date):
            if i in rwl_data[series]:
                series_data.append(rwl_data[series][i]/rwl_data[series]["div"])
            else:
                series_data.append(np.nan)
        df = pd.concat([df, pd.Series(data=series_data, name=series)], axis=1)
    return df

# Extract raw data from lines of .rwl file and store in a nested dictionary
def read_rwl(lines):
    rwl_data = {}
    first_date = sys.maxsize
    last_date = -sys.maxsize
    line_ct = 0
    for line in lines:
        try:
            line = line.rstrip("\n")

            if line[7] != '-' and line[6] != '-':
                series_id = line[:8].strip()
                iyr = int(line[8:12])
            elif line[7] == '-':
                series_id = line[:7].strip()
                iyr = int(line[7:12]) 
            elif line[6] == '-':
                series_id = line[:6].strip()
                iyr = int(line[6:12])

            if series_id not in rwl_data:
                rwl_data[series_id] = {}

            dataline = [line[i:i+6] for i in range(12, len(line), 6) if line[i:i+6].strip()]
            
            # keep track of the first and last date in the dataset
            line_start = int(iyr)
            first_date = min(first_date, line_start)
            last_date = max(last_date, (line_start+len(dataline)-1))

            # will implement some standardization here so that all data read is consistent, and all data written in rwl
            # can be written to one of the two popular precisions.
            for i in range(0, len(dataline)):
                if dataline[i].strip() == "999":
                    rwl_data[series_id]["div"] = 100
                    continue
                elif dataline[i].strip() == "-9999":
                    rwl_data[series_id]["div"] = 1000
                    continue
                data = float(int(dataline[i]))
                rwl_data[series_id][line_start+i] = data
            line_ct += 1

        except ValueError: # Stops reader, escalates to give the user an error when unexpected formatting is detected.
            print("Error reading line", line_ct + 1, ":\n", line, "\n")
            return None, None, None
    return rwl_data, first_date, last_date


# -*- coding: utf-8 -*-

__author__ = "Tyson Lee Swetnam"
__email__ = "tswetnam@arizona.edu"
__version__ = "0.1.5"

__copyright__ = """
   dplPy for tree ring width time series analyses
   Copyright (C) 2021  OpenDendro

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""
__license__ = "GNU GPLv3"


import os
import sys

lpath = os.path.dirname(os.path.realpath(__file__))
sys.path.append(lpath)

del os, sys


_hard_dependencies = ("pandas", "numpy", "scipy", "csaps", "matplotlib", "statsmodels")
_missing_dependencies = []

for _dependency in _hard_dependencies:
    try:
        __import__(_dependency)
    except ImportError as _e:  # pragma: no cover
        _missing_dependencies.append(f"{_dependency}: {_e}")

if _missing_dependencies:  # pragma: no cover
    raise ImportError(
        "Unable to import required dependencies:\n" + "\n".join(_missing_dependencies)
    )
del _hard_dependencies, _dependency, _missing_dependencies


from readers import readers
from summary import summary
from stats import stats
from report import report
from plot import plot
from detrend import detrend
from autoreg import ar_func, autoreg
from chron import chron
from chron_stabilized import chron_stabilized
from xdate import xdate, xdate_plot
from series_corr import series_corr
from writers import writers

__all__ = [
    readers,
    summary,
    stats,
    report,
    plot,
    detrend,
    ar_func,
    autoreg,
    chron,
    chron_stabilized,
    xdate,
    xdate_plot,
    series_corr,
    writers
]


from __future__ import print_function

__copyright__ = """
   dplPy for tree ring width time series analyses
   Copyright (C) 2024  OpenDendro

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""
__license__ = "GNU GPLv3"

#!/usr/bin/python
# -*- coding: utf-8 -*-

# Date: 9/11/2022
# Author: Ifeoluwa Ale
# Title: report.py
# Description: Generates a report about absent rings in the data set


import pandas as pd
from readers import readers
from stats import stats
import numpy as np
from statsmodels.tsa.ar_model import AutoReg

def report(inp: pd.DataFrame | str):
    """Generates a report
    
    Extended Summary
    ----------------
    Generates a text report about the input dataset that includes:
        Number of dated series
        Number of measurements
        Avg series length (years)
        Range (total years)
        Span (start-end year)
        Mean (Standard Deviation) series intercorrelation
        Mean (Standard Deviation) AR1
        Years with absent rings listed by series
    
    Parameters
    ----------
    data : str
           a data file (.CSV or .RWL) or a pandas dataframe imported from dpl.readers().
              
    Returns
    -------
    data : pandas dataframes
    
    Examples
    --------
    >>> import dplpy as dpl
    >>> data = dpl.readers("../tests/data/csv/file.csv")
    >>> dpl.report(data) 
    
    References
    ----------
    .. [1] https:/opendendro.org/dplpy-man/#report

    """
    if isinstance(inp, pd.DataFrame):
        series_data = inp
    elif isinstance(inp, str):
        series_data = readers(inp)
    else:
        return
    
    statistics = stats(series_data)
    no_of_series = series_data.shape[1]
    no_of_measurements = series_data.count().sum()
    first_year = statistics["first"].min()
    last_year = statistics["last"].max()
    missing_rings, avg_ar = get_report_stats(series_data)

    print("Number of dated series:", no_of_series)
    print("Number of measurements:", no_of_measurements)
    print("Avg series length:", round(no_of_measurements/no_of_series, 4))
    print("Range:", (last_year - first_year + 1))
    print("Span:", first_year, "-", last_year)
    print("Mean (Std dev) series intercorrelation:")
    print("Mean (Std dev) AR1:", round(avg_ar, 4))
    print("-------------")
    print("Years with absent rings listed by series\n")
    print_missing_ring_data(missing_rings)
    print("-------------")
    print("Years with internal NA values listed by series")

# Analyze the dataframe to generate report on missing data (and eventually internal NAs)
def get_report_stats(series_data):
    ar1s = []
    missing_rings = {}
    nans = {}
    for series_name, data in series_data.items():
        missing_rings[series_name] = list(map(str, data[data==0].index.tolist()))
        nans[series_name] = list(map(str, data[data==np.nan].index.tolist()))
        ar1s.append(round(AutoReg(data.dropna().to_numpy(), 1, old_names=False).fit().params[1], 3))
    avg_ar = sum(ar1s)/len(ar1s)

    #print(nans)
    internal_nans = {}
    for series_name, data in nans.items():
        if len(data) == 0:
            continue
        i = 1
        j = len(data) - 2
        
        while j > i:
            if data[i] != (data[i-1] + 1) and data[j+1] != (data[j] + 1):
                internal_nans[series_data] = data[i:j]
                break
            if data[i] == data[i-1] + 1:
                i += 1
            if data[j+1] == data[j] + 1:
                j += 1

    return missing_rings, avg_ar

# Print data about missing rings
def print_missing_ring_data(missing_rings):
    for series, missing in missing_rings.items():
        if len(missing) != 0:
            print("     ", series, "--", " ".join(missing))


from __future__ import print_function

__copyright__ = """
   dplPy for tree ring width time series analyses
   Copyright (C) 2021  OpenDendro

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""

__license__ = "GNU GPLv3"

#!/usr/bin/python
# -*- coding: utf-8 -*-

# Date: 11/17/2021 
# Author: Tyson Lee Swetnam
# Project: OpenDendro- Writers
# Description: Writers for all supported file types (*.CSV, *.RWL, and *.TXT)
# 
# example usages: 
# >>> import dplpy as dpl 
# >>> dpl.writers("./data/in_file.csv", "./data/out_file.rwl")
# >>> dpl.writers("./data/in_file.rwl", "./data/out_file.txt")
# >>> dpl.writers("./data/in_file.txt", "./data/out_file.csv")
# 
# example command line application:
# $ python src/dplpy.py writer --input ./data/in_file.csv --output ./data/out_file.rwl
#
# module uses two arguments: input file and output file 
# define `writer` module as a definition function
# input is expected to be a file path with file name and extension

import pandas as pd
import numpy as np
from chron import chron
from detrend import detrend

"""
    This function converts common ring width
    data files from one type to another
    It also allows you to append files that are missing metadata and write them back out
    Accepted file types are CSV, RWL, CRN (in dev) and TXT (in dev)
"""
def writers(data, label, format):
    if not isinstance(data, pd.DataFrame):
        raise TypeError("Expected input data to be pandas dataframe, not " + str(type(data)))
    
    if not isinstance(label, str):
        raise TypeError("Expected label to be of type str, not " + str(type(label)))
    
    if not isinstance(format, str):
        raise TypeError("Expected format to be of type str, not " + str(type(format)))
    
    filename = label + "." + format
    print("Writing to " + filename)
    output = open(filename, "w")
    if format == "csv":
        write_csv(data, output)
    elif format == "rwl":
        write_rwl(data, output)
    elif format == "crn":
        write_crn(data, label, output)
    elif format == "txt":
        write_txt(data, output)
    else:
        output.close()
        raise ValueError("Invalid file format given as parameter. Accepted file formats are csv, rwl, crn and txt")

    output.close()
    print("Done.")


def conv_data(data):
    if np.isnan(data):
        return "NA"
    else:
        return str(data)


def write_csv(data, file):
    file.write('"Year","')
    file.write('","'.join(data.columns.tolist()))
    file.write('"\n')

    for year, row in data.iterrows():
        file.write(str(year))
        file.write(",")
        file.write(",".join(map(conv_data, row)))
        file.write('\n')


# Incomplete. Doesn't account yet for varying precision standards for RWL files (lines ending in 999 vs -9999)
def write_rwl(data, file):
    for series in data.columns:
        start = data[series].first_valid_index()
        end = data[series].last_valid_index()
        i = start

        if i < 0:
            file.write(series.ljust(7))
            file.write(str(i).rjust(5))
        else:
            file.write(series.ljust(8))
            file.write(str(i).rjust(4))
        while i <= end:
            if np.isnan(data[series][i]):
                file.write(str(-9999).rjust(6))
                file.write("\n")
                while i <= end and np.isnan(data[series][i]):
                    i += 1
                if i <= end:
                    if i < 0:
                        file.write(series.ljust(7))
                        file.write(str(i).rjust(5))
                    else:
                        file.write(series.ljust(8))
                        file.write(str(i).rjust(4))
                continue

            file.write((f"{data[series][i]:.3f}").lstrip('0').replace('.', '').rjust(4, '0').rjust(6))
            i += 1
            if i % 10 == 0:
                file.write("\n")
                if i < 0:
                    file.write(series.ljust(7))
                    file.write(str(i).rjust(5))
                else:
                    file.write(series.ljust(8))
                    file.write(str(i).rjust(4))

        file.write(str(-9999).rjust(6))
        file.write("\n")


def write_crn(data, site_id, file, site_name="Unnamed object", species_code="UNKN", location="Unknown", species="Plantae", elevation="", lat_long="", investigator="", comp_date=""):
    rwi_data = chron(detrend(data, fit="spline", method="residual", plot=False), prewhiten=True)
    first = rwi_data.first_valid_index()
    last = rwi_data.last_valid_index()

    # For header, to be improved upon eventually
    file.write(site_id.ljust(9) + site_name.ljust(61) + species_code.ljust(4) + "\n")
    file.write(site_id.ljust(9) + location.ljust(13) + species.ljust(8) + elevation.ljust(5) + lat_long.ljust(10) + str(first).ljust(4) + str(last).rjust(5) + "\n")
    file.write(site_id.ljust(9) + investigator.ljust(73) + comp_date.ljust(8) + "\n")
    
    file.write(site_id.rjust(6))
    file.write(str(rwi_data.first_valid_index()).rjust(4))

    for year in rwi_data.index.to_numpy():
        file.write(str(round(rwi_data["Mean RWI"][year], 2)).replace(".", "").replace("0", "").rjust(4))
        file.write(str(rwi_data["Sample depth"][year]).rjust(3))
        
        if year % 10 == 9:
            # write TRL ID#(optional) which takes columns 82-88 
            if year + 1 in rwi_data.index.to_numpy():
                file.write("\n")
                file.write(site_id.rjust(6))
                file.write(str(year + 1).rjust(4))
            else:
                break
    file.write("9990  0")


def write_txt(data, file):
    header = ["year", "num".rjust(7), "seg".rjust(7), "age".rjust(7), "raw".rjust(7), "std".rjust(7), "res".rjust(7), "ars".rjust(7)]
    file.write("    ".join(header))
    file.write("\n")
    rwi_data = detrend(data, fit="spline", method="residual", plot=False)
    rwi_chron = chron(rwi_data, prewhiten=False)
    mean_res = chron(rwi_data, biweight=False, prewhiten=False)
    ar_chron = chron(rwi_data, prewhiten=True)

    first = rwi_chron.first_valid_index()
    last = rwi_chron.last_valid_index()

    for year in range(first, last+1):
        samp_dep = rwi_chron["Sample depth"][year]
        
        # standard chronology of detrended data
        std = rwi_chron["Mean RWI"][year]

        # residuals of detrended data?
        res =  mean_res["Mean RWI"][year]

        # residuals of ar modeled data?
        ars = ar_chron["Mean Res"][year]
        
        year_data = data.loc[[year]].dropna(axis=1)
        column_names = year_data.columns
        
        seg = 0
        age = 0
        raw = 0
        
        for name in column_names:
            seg += len(data[name].dropna())
            age += year - data[name].first_valid_index() + 1
            raw += data[name][year]
        
        seg = seg/len(column_names)
        age = age/len(column_names)
        raw = raw/len(column_names)

        

        # double check what res and ars are supposed to be
        # work on other columns
        line = [str(year).rjust(4), (f"{samp_dep:.3f}").rjust(7), (f"{seg:.3f}").rjust(7), (f"{age:.3f}").rjust(7), (f"{raw:.3f}").rjust(7), (f"{std:.3f}").rjust(7), (f"{res:.3f}").rjust(7), (f"{ars:.3f}").rjust(7),]
        file.write("    ".join(line))
        file.write("\n")



from __future__ import print_function

__copyright__ = """
   dplPy for tree ring width time series analyses
   Copyright (C) 2021  OpenDendro

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""
__license__ = "GNU GPL3"

#!/usr/bin/python
# -*- coding: utf-8 -*-

# Date:2021-08-13
# Author: Sarah Jackson
# Project: OpenDendro- Tucson Format 
# Description: Assigns variable names to each value in the tucson formated datasets using the emptyjson file.
# example usage:

import json

def create_json(input):

    with open("input", "r" ) as rings:
        data= rings.read()
        lines = data.split("\n")
        print (lines[0-2])
        # read every single line
        for  rows  in lines:
            print (rows)
        
        # remove empty lines from the file.

        # convert every element in each list to string- it is easier to manupilate the elements

        for items in lines:
            str(items)  

        # State/province
        state_province = lines[1,3]

        # country
        country= lines[0,2]

        # species
        species = lines[1,4]

        # species code
        species_code= lines[1,5]

        # start year- start of collection 
        start_date = int(lines[1,6])

        # end year- completion year
        end_year = int(lines[1,7])

    # latitude, longitude, elevation 
        latitude = lines[2,4]
        longitude = lines[2,5]

        # lead investigator
        lead_investigator= lines[2,2] 

        # site_id
            # Via indexing the first three digits of the site id are assigned to var named site_code and the rest to information.
            # Check if all the site codes are the same for the whole dataset 
        for site_code in lines:
            current=0
            information=0 
            site_id=0 
            unit = 0
            while current >=0 and current<=len(lines)-1:
                if current <=3:   
                    site_code = lines[current,0,0]       
                if current>=4 and current<=len(lines)-1:
                    if lines[current,0,0] == lines[4,0,0]:
                        site_id = lines[current,0,0]
                        information = site_id[3::]  
                site_id= site_code + information 
                #data
                data=[]
                y=1
                past= y-1
                end=[current,-1]
                if site_id == lines[current,0] and start_date <= int(lines[current]):
                    for points in lines[current,2:]:
                        pts=[]
                        pts.append(points)
                        # End of data collection for that year
                        if points== "999" or points== "-9999":
                            pts[:-1]
                            data.append(pts)
                            pts=0
                            unit = points
                else:
                    current = current + 1
                    pts=0
    
    jsonfile = {}
    jsonfile["site_code"] = site_code
    jsonfile["site_name"] = site_id
    jsonfile["species_code"] = species_code
    jsonfile["species_name"] = species
    jsonfile["country"] = country
    jsonfile["state_province"] = state_province
    jsonfile["elevation"] = ""
    jsonfile["latitude"] = latitude
    jsonfile["longitude"] = longitude
    if unit == "999":
        jsonfile["unit"] = "0.01mm"
    elif unit == "-9999":
        jsonfile["unit"] = "0.001mm"
    else:
        print("ERROR - Stop code is not 999 or -9999. Cannot identify a unit.")
        exit
    jsonfile["lead_investigator"] = lead_investigator
    jsonfile["collection_date"] = ""



__copyright__ = """
   dplPy for tree ring width time series analyses
   Copyright (C) 2024  OpenDendro

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

__license__ = "GNU GPLv3"

#!/usr/bin/python
# -*- coding: utf-8 -*-

# Date: 11/1/2022
# Author: Ifeoluwa Ale
# Title: tbrm.py
# Description: This file contains helper functions which find tukey's biweight robust mean of
#              an array like object.

import numpy as np

def tbrm(data, c=9):
    e = 1 * pow(10, -8)
    m = np.median(data)

    s = np.median(getabs(data - m)) # star?

    u = (data - m) / ((c * s) + e)

    w = np.zeros(len(data))

    for i in range(len(u)):
        if abs(u[i]) <= 1:
            w[i] = pow((1 - pow(u[i], 2)), 2)

    return np.sum(w*data)/np.sum(w)

def getabs(data):
    for i in range(len(data)):
        data[i] = abs(data[i])
    
    return data



__copyright__ = """
   dplPy for tree ring width time series analyses
   Copyright (C) 2024  OpenDendro

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""
__license__ = "GNU GPLv3"

#!/usr/bin/python
# -*- coding: utf-8 -*-

# Date: 5/27/2022
# Author: Ifeoluwa Ale
# Title: stats.py
# Description: Generates summary statistics for Tucson format and CSV format files
# example usage:
# >>> import dplpy as dpl 
# >>> data = dpl.readers("../tests/data/csv/file.csv")
# >>> dpl.stats(data)
#
# >>> dpl.stats("../tests/data/csv/file.csv")
# >>> Note: for file pathname inputs, only CSV and RWL file formats are accepted

# Create Summaries for Tucson (*rwl) files

# Ignore the following comments:
#Code to calculate ar1 when statsmodels can be imported
#from statsmodels.tsa import stattools
# x = 1-D array
# Yield normalized autocorrelation function of number lags
#autocorr = stattools.acf( x )

# Get autocorrelation coefficient at lag = 1
#autocorr_coeff = autocorr[1]

import pandas as pd
import numpy as np
from readers import readers
from statsmodels.tsa.ar_model import AutoReg

def stats(inp: pd.DataFrame | str):
    """Generates summary statistics
    
    Extended Summary
    ----------------
    Generates summary statistics for .RWL and .CSV format files. 
    It outputs a table with 'first', 'last', 'year', 'mean', 'median', 'stdev', 
    'skew', 'gini', 'ar1' for each series in data file.
    
    Parameters
    ----------
    data : str
           a data file (.CSV or .RWL) or a pandas dataframe imported from dpl.readers().
    
    Returns
    -------
    data : pandas dataframe
    
    Examples
    --------
    >>> dpl.stats(<data>)
    
    References
    ----------
    .. [1] https:/opendendro.org/dplpy-man/#stats
    
    
    """
    if isinstance(inp, pd.DataFrame):
        series_data = inp
    elif isinstance(inp, str):
        series_data = readers(inp)

        
    stats = {"series":[], "first":[], "last":[], "year": [], "mean": [], "median":[], "stdev":[], "skew":[], "gini":[], "ar1":[]}

    for series_name, data in series_data.items():
        stats["series"].append(series_name)
        stats["first"].append(data.first_valid_index())
        stats["last"].append(data.last_valid_index())
        stats["year"].append(stats["last"][-1] - stats["first"][-1] + 1)
        stats["mean"].append(round(data.mean(), 3))
        stats["median"].append(round(data.median(), 2))
        stats["stdev"].append(round(data.std(), 3))
        stats["skew"].append(round(get_skew(data), 3))
        stats["gini"].append(round(get_gini(data.dropna().to_numpy()), 3))
        stats["ar1"].append(round(AutoReg(data.dropna().to_numpy(), 1, old_names=False).fit().params[1], 3))


    statistics = pd.DataFrame(stats)
    statistics.index += 1
    return statistics

def get_gini(data_array):
    # might need to work on more efficient solution
    # Mean absolute difference
    mad = np.abs(np.subtract.outer(data_array, data_array)).mean()
    # Relative mean absolute difference
    rmad = mad/np.mean(data_array)
    # Gini coefficient
    g = 0.5 * rmad
    return g

# gets skew values for each series
def get_skew(data_series):
    return (((data_series - data_series.mean()) / data_series.std()) ** 3).mean()


__copyright__ = """
   dplPy for tree ring width time series analyses
   Copyright (C) 2024  OpenDendro

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

__license__ = "GNU GPLv3"

#!/usr/bin/python
# -*- coding: utf-8 -*-

# Date: 11/1/2022
# Author: Ifeoluwa Ale
# Title: smoothingspline.py
# Description: This contains the spline method which fits a series to
#              a spline curve.

from math import cos
from math import pi
from csaps import csaps

# Returns the spline parameter, given amplitude of the series and the period
def get_param(amp, period):
    freq = 1/period
    spline_param = 1/(((cos(2 * pi * freq) + 2) * (1 - amp)/(12 * amp * (cos(2 * pi * freq) - 1) ** 2))+ 1)
    return spline_param

def get_period(period, n):
    if period is None:
        return n * 0.67
    elif period < 0:
        return n * abs(period)/100  
    elif period <= 1:
        return n * period
    else:
        return period

# Fits a curve to the series given as input and returns the y-values of the curve
def spline(x, y, period=None):
    p = get_param(0.5, get_period(period, len(x)))
    yi = csaps(x, y, x, smooth=p)
    return yi


__copyright__ = """
   dplPy for tree ring width time series analyses
   Copyright (C) 2024  OpenDendro

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""
__license__ = "GNU GPLv3"

#!/usr/bin/python
# -*- coding: utf-8 -*-

# Date: 5/12/2023
# Author: Ifeoluwa Ale
# Title: rbar.py
# Description: Contains functions for finding best interval of overlapping series over a long
#              period of years, and calculating rbar constant for a dataset over this best
#              period of overlap

import numpy as np
from detrend import detrend
from chron import chron

# common_interval finds a range of years in the provided dataframe where there is maximum overlap between series over the longest period of years.
def common_interval(data):
    year = data.index.to_numpy() # this is the year vector
    crn = data.iloc[:,:] # these are the chronologies

    num_years = crn.shape[0]

    
    # across-column sum of non-NaN values to get the sample size = sample size
    sample_depth = np.sum(~np.isnan(crn), axis=1)
    # allocate
    N = np.full((num_years, num_years), np.nan) # square matrix with dimensions the length of the series (which reflects both starting year and possible block length)

    # loop over - this is a straight port from my MATLAB, possibly inefficient
    for i in range(num_years):  # effectively, looping over from 1 to the maximum length of the series in the data as potential lengths of a common interval block
        # define a block size, using smaller and smaller blocks as you get toward the last year of the series ... this loop therefore gets shorter as block size i gets larger ...
        for j in range(num_years - i):
            # for a starting year j and block length i, the smallest number of chronologies in that particular block
            N[j, i] = np.min(sample_depth[j:j+i+1])

    # pointwise multiplication of two square matrices - this essentially convolves sample size and block length to get number of pairwise comparisons possible
    N0 = N * np.tile(np.arange(num_years) + 1, (num_years, 1))

    # row (startyear) and column (block length) position of maximum value - is this an OK way to do this? tried other things that didn't work

    start_year, window_width = np.where(N0 == np.nanmax(N0))
    # this gives the same answer as MATLAB - 1828 to 1982 common interval
    return year[int(start_year)], year[int(start_year+window_width-1)]

  
# rbar returns a list of constants to multiply with each mean value generated for a range of years from a mean value chronology.
# Can use osborn, frank and 67spline methods to generate rbar values.
# Will be updated in the future to prioritize number of series, number of years or both. Currently attempts to do both.
def get_running_rbar(data, min_seg_ratio, method="osborn", corr_type="pearson"):
    # how we deal with nans will depend on method chosen for finding rbar. 
    # drop all series with nans for osborn, but drop only if they are not up to fraction of seg_length for frank

    # Osborn assumes all series are overlapping along the entire period. Drops none
    if method == "osborn":
        r_bar = mean_series_intercorrelation(data, corr_type, min_seg_ratio)
        return r_bar
    
    elif method == "frank":
        rel_data = data.copy()
        drop_columns = []

        # Identify columns that need to be dropped and drop them
        
        for column in rel_data:
            num_valid_elems = rel_data[column].size
            if num_valid_elems/data.shape[0] < min_seg_ratio:
                drop_columns.append(column)
        rel_data = rel_data.drop(columns=drop_columns)

        r_bar = mean_series_intercorrelation(rel_data, corr_type, min_seg_ratio)

        return r_bar
    
    elif method == "67spline":
        # probably need to update this
        rel_data = data.copy()
        signs = rel_data.where(rel_data < 0, 1)

        signs = signs.where(signs >= 0, -1)
        rel_series = rel_series.abs()
        rel_series_rwi = detrend(rel_series, fit="spline")
        res_frame = rel_series_rwi * signs
        return chron(res_frame, plot=False)['Mean RWI'].tolist()

    return None

def mean_series_intercorrelation(data_set, corr_type, min_seg_ratio):
    presence_df = data_set.notnull().astype('int')
    trans_presence_df = presence_df.transpose()

    corr_mat = data_set.corr(corr_type)
    np.fill_diagonal(corr_mat.values, np.nan)

    overlap_mat = trans_presence_df @ presence_df

    min_overlap = data_set.shape[0] * min_seg_ratio

    corr_mat.where(overlap_mat > min_overlap, inplace=True)
    
    return corr_mat.mean().mean()



from __future__ import print_function

__copyright__ = """
   dplPy for tree ring width time series analyses
   Copyright (C) 2024  OpenDendro

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""
__license__ = "GNU GPLv3"

#!/usr/bin/python
# -*- coding: utf-8 -*-

# Date: 
# Author: 
# Project: 
# Description: 

# example usages: 
# >>> 
# >>> 
#
# example command line application:
# $ 
# $

